@@ -38,13 +38,23 @@ impl<'mb> ModelBuilder<'mb> {
         self.scopes.push(HashMap::new());
         self.wire_body(&self.proto_model.doc.graph_def.body)?;
         let vars = self.scopes.pop().unwrap();
+
         let outputs = self
             .proto_model
             .doc
             .graph_def
             .results
             .iter()
-            .map(|s| vars[s].to::<OutletId>(self))
+            .map(|s| {
+                Ok(vars
+                    .get(s)
+                    .with_context(|| format!("Could not find variable for output named `{}'", s))?)
+            })
+            .collect::<TractResult<TVec<&Value>>>()?;
+
+        let outputs = outputs
+            .into_iter()
+            .map(|s| s.to::<OutletId>(self))
             .collect::<TractResult<TVec<OutletId>>>()?;
         self.model.set_output_outlets(&outputs)?;
         if let Some(properties) = self
@@ -57,7 +67,7 @@ impl<'mb> ModelBuilder<'mb> {
             .and_then(|body| body.get(0))
         {
             let properties: TVec<(String, Arc<Tensor>)> =
-                properties.right.resolve(self)?.to(self)?;
+                properties.right.resolve(self, &[])?.to(self)?;
             self.model.properties = properties.into_iter().collect();
         }
         Ok(())
@@ -74,32 +84,70 @@ impl<'mb> ModelBuilder<'mb> {
         // todo: can i relax the outlet id constraint ?
         for assignment in body {
             let identifiers = assignment.left.to_identifiers()?;
+            let datum_types = identifiers
+                .iter()
+                .map(|s| {
+                    self.proto_model
+                        .quantization
+                        .as_ref()
+                        .and_then(|qm| qm.get(*s).map(|q| q.datum_type()))
+                })
+                .collect::<Vec<_>>();
             self.naming_scopes.push(identifiers[0].to_string());
-            let values: TVec<OutletId> =
-                assignment.right.resolve(self).and_then(|v| v.to(self)).with_context(|| {
-                    format!("Plugging in assignement for {:?}", identifiers.join(", "))
-                })?;
-            if values.len() != identifiers.len() {
-                bail!(
-                    "Assignement for {} received {} value(s).",
-                    identifiers.join(","),
-                    values.len()
-                )
-            }
+            let values = if identifiers.len() == 1 {
+                let value: OutletId = assignment
+                    .right
+                    .resolve(self, &datum_types)
+                    .and_then(|v| v.to(self))
+                    .with_context(|| {
+                        format!("Plugging in assignement for {:?}", identifiers.join(", "))
+                    })?;
+                tvec!(value)
+            } else {
+                let values: TVec<OutletId> = assignment
+                    .right
+                    .resolve(self, &datum_types)
+                    .and_then(|v| v.to(self))
+                    .with_context(|| {
+                        format!("Plugging in assignement for {:?}", identifiers.join(", "))
+                    })?;
+                if values.len() != identifiers.len() {
+                    bail!(
+                        "Assignement for {} received {} value(s).",
+                        identifiers.join(","),
+                        values.len()
+                    )
+                }
+                values
+            };
             self.model.node_mut(values[0].node).name = format!("{}", self.naming_scopes.join("."));
             for (id, outlet) in identifiers.iter().zip(values.iter()) {
                 self.scopes.last_mut().unwrap().insert(id.to_string(), Value::Wire(*outlet));
             }
+            for (qparam, value) in datum_types.into_iter().zip(values.iter()) {
+                if let Some(qparam) = qparam {
+                    let node = self.model.node_mut(value.node);
+                    let output_fact = &mut node.outputs[value.slot].fact;
+                    output_fact.datum_type = qparam;
+                }
+            }
             self.naming_scopes.pop();
         }
         Ok(())
     }
 
-    pub fn wire_invocation(&mut self, invocation: &Invocation) -> TractResult<Value> {
+    pub fn wire_invocation(
+        &mut self,
+        invocation: &Invocation,
+        dt: &[Option<DatumType>],
+    ) -> TractResult<Value> {
         for frag in &self.proto_model.doc.fragments {
             if frag.decl.id == invocation.id && frag.body.is_some() {
-                let resolved =
-                    ResolvedInvocation { invocation, default_params: &frag.decl.parameters };
+                let resolved = ResolvedInvocation {
+                    invocation,
+                    dt_from_quant_file: dt,
+                    default_params: &frag.decl.parameters,
+                };
                 return self.wire_fragment_invocation(
                     &resolved,
                     &frag.decl,
@@ -109,7 +157,10 @@ impl<'mb> ModelBuilder<'mb> {
         }
         for registry in &self.framework.registries {
             if self.registries.contains(&registry.id) {
-                if let Some(outputs) = registry.deserialize(self, invocation)? {
+                if let Some(outputs) = registry
+                    .deserialize(self, invocation, dt)
+                    .with_context(|| format!("Interrogating registry {}", registry.id))?
+                {
                     return Ok(outputs);
                 }
             }
@@ -177,9 +228,10 @@ impl<'mb> ModelBuilder<'mb> {
     }
 }
 
-#[derive(Clone, Debug)]
+#[derive(Clone)]
 pub struct ResolvedInvocation<'a> {
     pub invocation: &'a Invocation,
+    pub dt_from_quant_file: &'a [Option<DatumType>],
     pub default_params: &'a [Parameter],
 }
 
@@ -190,7 +242,7 @@ impl<'a> ResolvedInvocation<'a> {
     {
         let rv = self.named_arg(name)?;
         let v = rv
-            .resolve(builder)
+            .resolve(builder, &[])
             .with_context(|| format!("Resolving argument `{}' ({:?})", name, rv))?;
         v.to::<T>(builder).with_context(|| format!("Converting argument `{}' from {:?}", name, v))
     }
@@ -248,19 +300,43 @@ impl LValue {
 impl Invocation {}
 
 impl RValue {
-    pub fn resolve(&self, builder: &mut ModelBuilder) -> TractResult<Value> {
+    pub fn resolve(
+        &self,
+        builder: &mut ModelBuilder,
+        dt: &[Option<DatumType>],
+    ) -> TractResult<Value> {
         match self {
             RValue::Identifier(id) => {
-                let outlet = builder
+                let mut outlet = builder
                     .scopes
                     .last()
                     .unwrap()
                     .get(id)
                     .cloned()
                     .ok_or_else(|| format_err!("No value for name {}", id))?;
+                if let Value::Wire(outlet_id) = outlet {
+                    let out_dt =
+                        builder.model.node(outlet_id.node).outputs[outlet_id.slot].fact.datum_type;
+                    if let Some(Some(dt)) = dt.get(0) {
+                        if out_dt.unquantized() != dt.unquantized() {
+                            return Err(format_err!(
+                                "Mismatched types expected {:?}, got {:?}",
+                                dt,
+                                out_dt
+                            ));
+                        }
+                        if out_dt != *dt {
+                            outlet = Value::Wire(
+                                builder.wire(tract_core::ops::cast::cast(*dt), &[outlet_id])?[0],
+                            )
+                        }
+                    }
+                }
                 Ok(outlet)
             }
-            RValue::Invocation(inv) => builder.wire_invocation(inv),
+            RValue::Invocation(inv) => builder
+                .wire_invocation(inv, dt)
+                .with_context(|| format!("Resolving invocation {}", inv.id)),
             RValue::Binary(left, op, right) => {
                 let op = match &**op {
                     "+" => "add",
@@ -284,14 +360,40 @@ impl RValue {
                         Argument { id: None, rvalue: right.as_ref().clone() },
                     ],
                 };
-                builder.wire_invocation(&inv)
+                builder
+                    .wire_invocation(&inv, dt)
+                    .with_context(|| format!("Resolving invocation {}", inv.id))
             }
             RValue::Array(array) => Ok(Value::Array(
-                array.iter().map(|i| i.resolve(builder)).collect::<TractResult<_>>()?,
-            )),
-            RValue::Tuple(array) => Ok(Value::Tuple(
-                array.iter().map(|i| i.resolve(builder)).collect::<TractResult<_>>()?,
+                array
+                    .iter()
+                    .zip(std::iter::repeat(&dt.get(0).copied().flatten()))
+                    .map(|(i, dt)| i.resolve(builder, &[*dt]))
+                    .collect::<TractResult<_>>()?,
             )),
+            RValue::Tuple(array) => {
+                let dt_iter: Box<dyn Iterator<Item = &Option<DatumType>>> =
+                    if dt.len() == 0 || dt.len() == 1 && dt[0] == None {
+                        Box::new(std::iter::repeat(&None))
+                    } else if dt.len() == array.len() {
+                        Box::new(dt.iter())
+                    } else {
+                        bail!("Wrong number of types for a tuple, got {:?} for {:?}", dt, array)
+                    };
+                Ok(Value::Tuple(
+                    array
+                        .iter()
+                        .zip(dt_iter)
+                        .map(|(i, dt)| {
+                            if dt.is_none() {
+                                i.resolve(builder, &[])
+                            } else {
+                                i.resolve(builder, &[*dt])
+                            }
+                        })
+                        .collect::<TractResult<_>>()?,
+                ))
+            }
             RValue::Literal(Literal::Numeric(f)) => {
                 if f.contains(".") || f.contains("e") {
                     f.parse::<f32>()
@@ -308,7 +410,8 @@ impl RValue {
             RValue::Literal(Literal::Array(array)) => Ok(Value::Array(
                 array
                     .iter()
-                    .map(|i| RValue::Literal(i.clone()).resolve(builder))
+                    .zip(std::iter::repeat(&dt.get(0).copied().flatten()))
+                    .map(|(i, dt)| RValue::Literal(i.clone()).resolve(builder, &[*dt]))
                     .collect::<TractResult<_>>()?,
             )),
             _ => panic!("{:?}", self),
@@ -352,6 +455,7 @@ impl CoerceFrom<Value> for Value {
 impl CoerceFrom<Value> for Arc<Tensor> {
     fn coerce(builder: &mut ModelBuilder, from: &Value) -> TractResult<Self> {
         match from {
+            Value::Dim(t) => Ok(rctensor0(t.to_i32()?)),
             Value::Tensor(t) => Ok(t.clone()),
             Value::Scalar(f) => Ok(rctensor0(*f)),
             Value::String(f) => Ok(rctensor0(f.clone())),
@@ -361,6 +465,35 @@ impl CoerceFrom<Value> for Arc<Tensor> {
                 .konst
                 .clone()
                 .ok_or_else(|| format_err!("Not a const")),
+            Value::Array(items) => {
+                let mut tensors = vec![];
+                for item in items {
+                    let tensor = Arc::<Tensor>::coerce(builder, item)?;
+                    let mut tensor = tensor.into_tensor();
+                    tensor.insert_axis(0)?;
+                    tensors.push(tensor);
+                }
+                let tensor = Tensor::stack_tensors(0, &tensors)?;
+                Ok(tensor.into_arc_tensor())
+            }
+            _ => bail!("Can not build a tensor from {:?}", from),
+        }
+    }
+}
+
+impl CoerceFrom<Value> for (Arc<Tensor>, DatumType) {
+    fn coerce(builder: &mut ModelBuilder, from: &Value) -> TractResult<Self> {
+        match from {
+            Value::Tensor(t) => Ok((t.clone(), t.datum_type())),
+            Value::Scalar(f) => Ok((rctensor0(*f), DatumType::F32)),
+            Value::String(f) => Ok((rctensor0(f.clone()), DatumType::String)),
+            Value::Wire(o) => {
+                let outlet_fact = builder.model.outlet_fact(*o)?;
+                Ok((
+                    outlet_fact.konst.clone().ok_or_else(|| format_err!("Not a const"))?,
+                    outlet_fact.datum_type,
+                ))
+            }
             _ => bail!("Can not build a tensor from {:?}", from),
         }
     }
@@ -369,6 +502,9 @@ impl CoerceFrom<Value> for Arc<Tensor> {
 impl CoerceFrom<Value> for OutletId {
     fn coerce(builder: &mut ModelBuilder, from: &Value) -> TractResult<Self> {
         match from {
+            Value::Tensor(t) => {
+                Ok(builder.wire(tract_core::ops::konst::Const::new(t.clone()), &[])?[0])
+            }
             Value::Scalar(f) => {
                 Ok(builder.wire(tract_core::ops::konst::Const::new(rctensor0(*f)), &[])?[0])
             }
@@ -377,6 +513,10 @@ impl CoerceFrom<Value> for OutletId {
             }
             Value::Wire(outlet) => Ok(*outlet),
             Value::Tuple(tuple) if tuple.len() == 1 => OutletId::coerce(builder, &tuple[0]),
+            Value::Array(_) => {
+                let tensor = Arc::<Tensor>::coerce(builder, from)?;
+                Ok(builder.wire(tract_core::ops::konst::Const::new(tensor), &[])?[0])
+            }
             _ => bail!("Can not build an outletid from {:?}", from),
         }
     }
